@@ -4,35 +4,22 @@
 
 import errno
 import os
-import string
-import warnings
 
-from pyfatfs import FAT_OEM_ENCODING
+from pyfatfs import FAT_OEM_ENCODING, _init_check
 from pyfatfs._exceptions import PyFATException, NotAFatEntryException
 
 
-def _init_check(func):
-    def _wrapper(*args, **kwargs):
-        initialised = args[0].initialised
-
-        if initialised is True:
-            return func(*args, **kwargs)
-        else:
-            raise PyFATException("Class has not yet been fully initialised, "
-                                 "please instantiate first.")
-
-    return _wrapper
-
-
 class EightDotThree:
     """8DOT3 filename representation."""
 
     #: Length of the byte representation in a directory entry header
     SFN_LENGTH = 11
 
-    #: Valid characters for 8.3 file names
-    VALID_CHARACTERS = f"{string.ascii_letters.upper()}{string.digits}" \
-                       "!#$%&'()-@^_`{}~"
+    #: Invalid characters for 8.3 file names
+    INVALID_CHARACTERS = [range(0x0, 0x20)] + [0x22, 0x2A, 0x2B, 0x2C, 0x2E,
+                                               0x2F, 0x3A, 0x3B, 0x3C, 0x3D,
+                                               0x3E, 0x3F, 0x5B, 0x5C, 0x5D,
+                                               0x7C]
 
     def __init__(self, encoding: str = FAT_OEM_ENCODING):
         """Offer 8DOT3 filename operation.
@@ -40,28 +27,31 @@ class EightDotThree:
         :param encoding: Codepage for the 8.3 filename.
                          Defaults to `FAT_OEM_ENCODING` as per FAT spec.
         """
-        self.name: bytes = None
+        self.name: bytearray = None
         self.encoding = encoding
-        self.initialised = False
+        self.initialized = False
+
+    def __str__(self):
+        """Decode and un-pad SFN string."""
+        name = self.name
+        if name[0] == 0x05:
+            # Translate 0x05 to 0xE5
+            name[0] = 0xE5
+
+        base = name[:8].decode(self.encoding).rstrip()
+        ext = name[8:11].decode(self.encoding).rstrip()
+        sep = "." if len(ext) > 0 else ""
+
+        return sep.join([base, ext])
 
     def __bytes__(self):
         """Byte representation of the 8DOT3 name dir entry headers."""
-        return self.name
-
-    def byte_repr(self):
-        """Do not use."""
-        warnings.warn("byte_repr is deprecated, directly cast "
-                      "to bytes instead", DeprecationWarning)
-        return bytes(self)
+        return bytes(self.name)
 
     @_init_check
     def get_unpadded_filename(self) -> str:
         """Retrieve the human readable filename."""
-        base = self.name[:8].decode(self.encoding).rstrip()
-        ext = self.name[8:11].decode(self.encoding).rstrip()
-        sep = "." if len(ext) > 0 else ""
-
-        return sep.join([base, ext])
+        return str(self)
 
     @staticmethod
     def __raise_8dot3_nonconformant(name: str):
@@ -72,12 +62,11 @@ class EightDotThree:
 
     def __set_name(self, name: bytes):
         """Set self.name and verify for correctness."""
-        name_str = name.decode(self.encoding)
-        if len(name_str) != 11:
-            self.__raise_8dot3_nonconformant(name_str)
+        if len(name) != 11:
+            self.__raise_8dot3_nonconformant(name.decode(self.encoding))
 
         self.name = name
-        self.initialised = True
+        self.initialized = True
 
     def set_byte_name(self, name: bytes):
         """Set the name as byte input from a directory entry header.
@@ -88,6 +77,8 @@ class EightDotThree:
             raise TypeError(f"Given parameter must be of type bytes,"
                             f"but got {type(name)} instead.")
 
+        name = bytearray(name)
+
         if len(name) != 11:
             raise ValueError("Invalid byte name supplied, must be exactly "
                              "11 bytes long (8+3).")
@@ -97,10 +88,6 @@ class EightDotThree:
             raise NotAFatEntryException("Given dir entry is invalid and has "
                                         "no valid name.", free_type=name[0])
 
-        if name[0] == 0x05:
-            # Translate 0x05 to 0xE5
-            name = name.replace(bytes(0x05), bytes(0xE5), 1)
-
         self.__set_name(name)
 
     def set_str_name(self, name: str):
@@ -109,11 +96,14 @@ class EightDotThree:
             raise TypeError(f"Given parameter must be of type str,"
                             f"but got {type(name)} instead.")
 
-        if not self.is_8dot3_conform(name):
+        if not self.is_8dot3_conform(name, self.encoding):
             self.__raise_8dot3_nonconformant(name)
 
-        name = self._pad_8dot3_name(name)
-        self.__set_name(name.encode(self.encoding))
+        name = bytearray(self._pad_8dot3_name(name).encode(self.encoding))
+        if name[0] == 0xE5:
+            name[0] = 0x05
+        self.name = name
+        self.initialized = True
 
     @_init_check
     def checksum(self) -> int:
@@ -128,25 +118,25 @@ class EightDotThree:
         return chksum
 
     @staticmethod
-    def __check_characters(name: str) -> bool:
+    def __check_characters(name: str, encoding: str) -> bool:
         """Test if given string contains invalid chars for 8.3 names.
 
         :param name: `str`: Filename to parse
         :raises: `ValueError` if the given string contains invalid
                  8.3 filename characters.
         """
-        vc = EightDotThree.VALID_CHARACTERS
-        n = ''.join(filter(lambda x: x in vc, name))
-        if n != name:
-            raise ValueError(f"Invalid characters in string '{name}', "
-                             f"cannot be used as part of an 8.3 "
-                             f"conform file name.")
+        for c in name:
+            if c.encode(encoding) in EightDotThree.INVALID_CHARACTERS:
+                raise ValueError(f"Invalid characters in string '{name}', "
+                                 f"cannot be used as part of an 8.3 "
+                                 f"conform file name.")
 
     @staticmethod
-    def is_8dot3_conform(entry_name: str):
+    def is_8dot3_conform(entry_name: str, encoding: str = FAT_OEM_ENCODING):
         """Indicate conformance of given entries name to 8.3 standard.
 
         :param entry_name: Name of entry to check
+        :param encoding: ``str``: Encoding for SFN
         :returns: bool indicating conformance of name to 8.3 standard
         """
         if entry_name != entry_name.upper():
@@ -163,7 +153,7 @@ class EightDotThree:
         # Check for valid characters in both filename segments
         for i in [root, ext]:
             try:
-                EightDotThree.__check_characters(i)
+                EightDotThree.__check_characters(i, encoding=encoding)
             except ValueError:
                 return False
 
@@ -213,7 +203,7 @@ class EightDotThree:
             char = char.upper()
             if char == ' ':
                 return ''
-            if char not in EightDotThree.VALID_CHARACTERS:
+            if char in EightDotThree.INVALID_CHARACTERS:
                 return '_'
             return char
 
