@@ -3,7 +3,8 @@
 """Directory entry operations with PyFAT."""
 import posixpath
 import struct
-import warnings
+
+from time import timezone
 
 from pyfatfs.DosDateTime import DosDateTime
 from pyfatfs.EightDotThree import EightDotThree
@@ -17,6 +18,11 @@ import errno
 class FATDirectoryEntry:
     """Represents directory entries in FAT (files & directories)."""
 
+    #: Marks a directory entry as empty
+    FREE_DIR_ENTRY_MARK = 0xE5
+    #: Marks all directory entries after this one as empty
+    LAST_DIR_ENTRY_MARK = 0x00
+
     #: Bit set in DIR_Attr if entry is read-only
     ATTR_READ_ONLY = 0x01
     #: Bit set in DIR_Attr if entry is hidden
@@ -48,11 +54,12 @@ class FATDirectoryEntry:
                           "DIR_WrtDate", "DIR_FstClusLO",
                           "DIR_FileSize"]
 
-    def __init__(self, DIR_Name, DIR_Attr, DIR_NTRes, DIR_CrtTimeTenth,
-                 DIR_CrtTime, DIR_CrtDate, DIR_LstAccessDate,
-                 DIR_FstClusHI, DIR_WrtTime, DIR_WrtDate, DIR_FstClusLO,
-                 DIR_FileSize, encoding: str = FAT_OEM_ENCODING,
-                 lfn_entry=None):
+    def __init__(self, DIR_Name: EightDotThree, DIR_Attr: int,
+                 DIR_NTRes: int, DIR_CrtTimeTenth: int,
+                 DIR_CrtTime: int, DIR_CrtDate: int, DIR_LstAccessDate: int,
+                 DIR_FstClusHI: int, DIR_WrtTime: int, DIR_WrtDate: int,
+                 DIR_FstClusLO: int, DIR_FileSize: int,
+                 encoding: str = FAT_OEM_ENCODING, lfn_entry=None):
         """FAT directory entry constructor.
 
         :param DIR_Name: `EightDotThree` class instance
@@ -70,7 +77,7 @@ class FATDirectoryEntry:
         :param encoding: Encoding of filename
         :param lfn_entry: FATLongDirectoryEntry instance or None
         """
-        self.name = DIR_Name
+        self.name: EightDotThree = DIR_Name
         self.attr = int(DIR_Attr)
         self.ntres = int(DIR_NTRes)
         self.crttimetenth = int(DIR_CrtTimeTenth)
@@ -92,6 +99,40 @@ class FATDirectoryEntry:
         self.__dirs = set()
         self.__encoding = encoding
 
+    @staticmethod
+    def new(name: EightDotThree, tz: timezone, encoding: str,
+            attr: int = 0, ntres: int = 0, cluster: int = 0,
+            filesize: int = 0) -> "FATDirectoryEntry":
+        """Create a new directory entry with sane defaults.
+
+        :param name: ``EightDotThree``: SFN of new dentry
+        :param tz: ``timezone``: Timezone value to use for new timestamp
+        :param encoding: ``str``: Encoding for SFN
+        :param attr: ``int``: Directory attributes
+        :param ntres: ``int``: Reserved NT directory attributes
+        :param cluster: ``int``: Cluster number of dentry
+        :param filesize: ``int``: Size of file referenced by dentry
+        :returns: ``FATDirectoryEntry`` instance
+        """
+        dt = DosDateTime.now(tz=tz)
+        dentry = FATDirectoryEntry(
+            DIR_Name=name,
+            DIR_Attr=attr,
+            DIR_NTRes=ntres,
+            DIR_CrtTimeTenth=0,
+            DIR_CrtTime=dt.serialize_time(),
+            DIR_CrtDate=dt.serialize_date(),
+            DIR_LstAccessDate=dt.serialize_date(),
+            DIR_FstClusHI=0x00,
+            DIR_WrtTime=dt.serialize_time(),
+            DIR_WrtDate=dt.serialize_date(),
+            DIR_FstClusLO=0x00,
+            DIR_FileSize=filesize,
+            encoding=encoding
+        )
+        dentry.set_cluster(cluster)
+        return dentry
+
     def get_ctime(self) -> DosDateTime:
         """Get dentry creation time."""
         return self.__combine_dosdatetime(self.crtdate, self.crttime)
@@ -145,6 +186,8 @@ class FATDirectoryEntry:
         sz = self.FAT_DIRECTORY_HEADER_SIZE
         if isinstance(self.lfn_entry, FATLongDirectoryEntry):
             sz *= len(self.lfn_entry.lfn_entries)
+        sz += self.FAT_DIRECTORY_HEADER_SIZE * len(self.__dirs)+1
+
         return sz
 
     def get_size(self):
@@ -152,11 +195,6 @@ class FATDirectoryEntry:
 
         :returns: Filesize or directory entry size in bytes as int
         """
-        if self.is_directory():
-            sz = self.FAT_DIRECTORY_HEADER_SIZE
-            sz *= len(self.__dirs)+1
-            return sz
-
         return self.filesize
 
     def set_size(self, size: int):
@@ -185,28 +223,19 @@ class FATDirectoryEntry:
 
         :returns: Entry & LFN entry as bytes-object
         """
-        name = bytes(self.name)
-        if name[0] == 0xE5:
-            name[0] = 0x05
-
         entry = b''
         if isinstance(self.lfn_entry, FATLongDirectoryEntry):
             entry += bytes(self.lfn_entry)
 
-        entry += struct.pack(self.FAT_DIRECTORY_LAYOUT, name, self.attr,
-                             self.ntres, self.crttimetenth, self.crttime,
-                             self.crtdate, self.lstaccessdate,
+        entry += struct.pack(self.FAT_DIRECTORY_LAYOUT,
+                             self.name.name,
+                             self.attr, self.ntres, self.crttimetenth,
+                             self.crttime, self.crtdate, self.lstaccessdate,
                              self.fstclushi, self.wrttime, self.wrtdate,
                              self.fstcluslo, self.filesize)
 
         return entry
 
-    def byte_repr(self):
-        """Do not use."""
-        warnings.warn("byte_repr is deprecated, directly cast "
-                      "to bytes instead", DeprecationWarning)
-        return bytes(self)
-
     def _add_parent(self, cls):
         """Add parent directory link to current directory entry.
 
@@ -241,7 +270,7 @@ class FATDirectoryEntry:
         if self._parent is None:
             return "/"
 
-        return posixpath.join(list(reversed(
+        return posixpath.join(*list(reversed(
             self._parent._get_parent_dir(parent_dirs))))
 
     def get_parent_dir(self):
@@ -416,12 +445,21 @@ class FATDirectoryEntry:
         dir_entry._add_parent(self)
         self.__dirs.add(dir_entry)
 
+    def mark_empty(self):
+        """Mark this directory entry as empty."""
+        # Also mark LFN entries as empty
+        try:
+            self.lfn_entry.mark_empty()
+        except AttributeError:
+            pass
+
+        self.name.name[0] = self.FREE_DIR_ENTRY_MARK
+
     def remove_dir_entry(self, name):
         """Remove given dir_entry from dir list.
 
         **NOTE:** This will also remove special entries such
         as ».«, »..« and volume labels!
-
         """
         d, f, s = self.get_entries()
 
@@ -440,7 +478,7 @@ class FATDirectoryEntry:
                              f"file or directory!", errno=errno.ENOENT)
 
     def __repr__(self):
-        """String-represent directory entry by (preferrably) LFN.
+        """String-represent directory entry by (preferably) LFN.
 
         :returns: str: Long file name if existing, 8DOT3 otherwise
         """
@@ -498,6 +536,12 @@ class FATLongDirectoryEntry(object):
                            reverse=reverse):
             yield e
 
+    def mark_empty(self):
+        """Mark LFN entry as empty."""
+        free_dir_entry_mark = FATDirectoryEntry.FREE_DIR_ENTRY_MARK
+        for k in self.lfn_entries.keys():
+            self.lfn_entries[k]["LDIR_Ord"] = free_dir_entry_mark
+
     def __bytes__(self):
         """Represent LFN entries as bytes."""
         entries_bytes = b""
@@ -509,12 +553,6 @@ class FATLongDirectoryEntry(object):
                                          e["LDIR_FstClusLO"], e["LDIR_Name3"])
         return entries_bytes
 
-    def byte_repr(self):
-        """Do not use."""
-        warnings.warn("byte_repr is deprecated, directly cast "
-                      "to bytes instead", DeprecationWarning)
-        return bytes(self)
-
     def __str__(self):
         """Remove padding from LFN entry and decode it.
 
@@ -548,7 +586,8 @@ class FATLongDirectoryEntry(object):
         lfn_attr_mask = FATDirectoryEntry.ATTR_LONG_NAME_MASK
         is_attr_set = (LDIR_Attr & lfn_attr_mask) == lfn_attr
 
-        return is_attr_set and LDIR_Ord != 0xE5
+        return is_attr_set and \
+            LDIR_Ord != FATDirectoryEntry.FREE_DIR_ENTRY_MARK
 
     def add_lfn_entry(self, LDIR_Ord, LDIR_Name1, LDIR_Attr, LDIR_Type,
                       LDIR_Chksum, LDIR_Name2, LDIR_FstClusLO, LDIR_Name3):
@@ -601,7 +640,7 @@ class FATLongDirectoryEntry(object):
 
 
 def make_lfn_entry(dir_name: str,
-                   short_name):
+                   short_name: EightDotThree):
     """Generate a `FATLongDirectoryEntry` instance from directory name.
 
     :param dir_name: Long name of directory
@@ -616,7 +655,8 @@ def make_lfn_entry(dir_name: str,
     dir_name = dir_name.encode(FAT_LFN_ENCODING)
     dir_name_modulus = len(dir_name) % lfn_entry_length
 
-    if EightDotThree.is_8dot3_conform(dir_name_str):
+    if EightDotThree.is_8dot3_conform(dir_name_str,
+                                      encoding=short_name.encoding):
         raise PyFATException("Directory entry is already 8.3 conform, "
                              "no need to create an LFN entry.",
                              errno=errno.EINVAL)
