@@ -4,6 +4,9 @@
 import datetime
 import posixpath
 import errno
+from copy import copy
+from io import BytesIO, IOBase
+from typing import Union
 
 from fs.base import FS
 from fs.mode import Mode
@@ -174,7 +177,7 @@ class PyFatFS(FS):
         basename = "/".join(path.split("/")[:-1])
         dirname = path.split("/")[-1]
 
-        # Plausability checks
+        # Plausibility checks
         try:
             self.opendir(basename)
         except DirectoryExpected:
@@ -186,11 +189,16 @@ class PyFatFS(FS):
         except ResourceNotFound:
             pass
         else:
+            if dentry.is_directory():
+                raise FileExpected(path)
             if not wipe:
                 return False
             else:
                 # Clean up existing file contents
-                # TODO: touch {a,m}time
+                dt = DosDateTime.now(tz=self.tz)
+                dentry.wrttime = dt.serialize_time()
+                dentry.wrtdate = dt.serialize_date()
+                dentry.lstaccessdate = dt.serialize_date()
                 dentry.set_size(0)
                 old_cluster = dentry.get_cluster()
                 dentry.set_cluster(0)
@@ -198,25 +206,12 @@ class PyFatFS(FS):
                 return True
 
         # Determine 8DOT3 file name + LFN
-        short_name = EightDotThree()
+        short_name = EightDotThree(encoding=self.fs.encoding)
         n = short_name.make_8dot3_name(dirname, base)
         short_name.set_str_name(n)
 
-        dt = DosDateTime.now(tz=self.tz)
-
-        newdir = FATDirectoryEntry(DIR_Name=short_name,
-                                   DIR_Attr=0,
-                                   DIR_NTRes=0,
-                                   DIR_CrtTimeTenth=0,
-                                   DIR_CrtTime=dt.serialize_time(),
-                                   DIR_CrtDate=dt.serialize_date(),
-                                   DIR_LstAccessDate=dt.serialize_date(),
-                                   DIR_FstClusHI=0x00,
-                                   DIR_WrtTime=dt.serialize_time(),
-                                   DIR_WrtDate=dt.serialize_date(),
-                                   DIR_FstClusLO=0x00,
-                                   DIR_FileSize=0,
-                                   encoding=self.fs.encoding)
+        newdir = FATDirectoryEntry.new(name=short_name, tz=self.tz,
+                                       encoding=self.fs.encoding)
 
         # Create LFN entry if required
         _sfn = short_name.get_unpadded_filename()
@@ -244,7 +239,7 @@ class PyFatFS(FS):
         base = split(path)[0]
         dirname = split(path)[1]
 
-        # Plausability checks
+        # Plausibility checks
         try:
             self.opendir(base)
         except DirectoryExpected:
@@ -252,38 +247,25 @@ class PyFatFS(FS):
         base = self._get_dir_entry(base)
 
         try:
-            self._get_dir_entry(path)
+            dentry = self._get_dir_entry(path)
         except ResourceNotFound:
             pass
         else:
-            if not recreate:
+            if not recreate or not dentry.is_directory():
                 raise DirectoryExists(path)
             else:
-                # TODO: Update mtime
                 return SubFS(self, path)
 
         parent_is_root = base == self.fs.root_dir
 
         # Determine 8DOT3 file name + LFN
-        short_name = EightDotThree()
+        short_name = EightDotThree(encoding=self.fs.encoding)
         n = short_name.make_8dot3_name(dirname, base)
         short_name.set_str_name(n)
 
-        dt = DosDateTime.now(tz=self.tz)
-
-        newdir = FATDirectoryEntry(DIR_Name=short_name,
-                                   DIR_Attr=FATDirectoryEntry.ATTR_DIRECTORY,
-                                   DIR_NTRes=0,
-                                   DIR_CrtTimeTenth=0,
-                                   DIR_CrtTime=dt.serialize_time(),
-                                   DIR_CrtDate=dt.serialize_date(),
-                                   DIR_LstAccessDate=dt.serialize_date(),
-                                   DIR_FstClusHI=0x00,
-                                   DIR_WrtTime=dt.serialize_time(),
-                                   DIR_WrtDate=dt.serialize_date(),
-                                   DIR_FstClusLO=0x00,
-                                   DIR_FileSize=0,
-                                   encoding=self.fs.encoding)
+        newdir = FATDirectoryEntry.new(name=short_name, tz=self.tz,
+                                       attr=FATDirectoryEntry.ATTR_DIRECTORY,
+                                       encoding=self.fs.encoding)
 
         # Create LFN entry if required
         _sfn = short_name.get_unpadded_filename()
@@ -298,36 +280,19 @@ class PyFatFS(FS):
         newdir.set_cluster(first_cluster)
         dot_sn = EightDotThree()
         dot_sn.set_byte_name(b".          ")
-        dot = FATDirectoryEntry(DIR_Name=dot_sn,
-                                DIR_Attr=FATDirectoryEntry.ATTR_DIRECTORY,
-                                DIR_NTRes=newdir.ntres,
-                                DIR_CrtTimeTenth=newdir.crttimetenth,
-                                DIR_CrtTime=newdir.crttime,
-                                DIR_CrtDate=newdir.crtdate,
-                                DIR_LstAccessDate=newdir.lstaccessdate,
-                                DIR_FstClusHI=newdir.fstclushi,
-                                DIR_WrtTime=newdir.wrttime,
-                                DIR_WrtDate=newdir.wrtdate,
-                                DIR_FstClusLO=newdir.fstcluslo,
-                                DIR_FileSize=newdir.filesize,
-                                encoding=self.fs.encoding)
+        dot = copy(newdir)
+        dot.name = dot_sn
+        dot.lfn_entry = None
+        dot._parent = None
+
         dotdot_sn = EightDotThree()
         dotdot_sn.set_byte_name(b"..         ")
-        base_fstclushi = base.fstclushi if not parent_is_root else 0x0
-        base_fstcluslo = base.fstcluslo if not parent_is_root else 0x0
-        dotdot = FATDirectoryEntry(DIR_Name=dotdot_sn,
-                                   DIR_Attr=FATDirectoryEntry.ATTR_DIRECTORY,
-                                   DIR_NTRes=base.ntres,
-                                   DIR_CrtTimeTenth=base.crttimetenth,
-                                   DIR_CrtTime=base.crttime,
-                                   DIR_CrtDate=base.crtdate,
-                                   DIR_LstAccessDate=base.lstaccessdate,
-                                   DIR_FstClusHI=base_fstclushi,
-                                   DIR_WrtTime=base.wrttime,
-                                   DIR_WrtDate=base.wrtdate,
-                                   DIR_FstClusLO=base_fstcluslo,
-                                   DIR_FileSize=base.filesize,
-                                   encoding=self.fs.encoding)
+        dotdot = copy(base)
+        dotdot.name = dotdot_sn
+        dotdot.lfn_entry = None
+        dotdot._parent = None
+        if parent_is_root:
+            dotdot.set_cluster(0)
         newdir.add_subdirectory(dot)
         newdir.add_subdirectory(dotdot)
 
@@ -423,8 +388,13 @@ class PyFatFS(FS):
         parent_dir.remove_dir_entry(str(dir_entry))
         self.fs.update_directory_entry(parent_dir)
 
+        # Mark dentry as free
+        dir_entry.mark_empty()
+        if dir_entry.is_directory():
+            self.fs.update_directory_entry(dir_entry)
+
         # Free cluster in FAT
-        if dir_entry.get_size() > 0 and dir_entry.get_cluster() != 0:
+        if dir_entry.get_entry_size() > 0 and dir_entry.get_cluster() != 0:
             # Empty files have a cluster ID of 0
             self.fs.free_cluster_chain(dir_entry.get_cluster())
         del dir_entry
@@ -513,3 +483,37 @@ class PyFatFS(FS):
             dentry.lstaccessdate = atime.serialize_date()
 
         self.fs.update_directory_entry(dentry.get_parent_dir())
+
+
+class PyFatBytesIOFS(PyFatFS):
+    """Provide PyFatFS functionality for BytesIO or IOBase streams."""
+
+    def __init__(self, fp: Union[IOBase, BytesIO],
+                 encoding: str = FAT_OEM_ENCODING,
+                 offset: int = 0, preserve_case: bool = True,
+                 utc: bool = False):
+        """PyFilesystem2 FAT constructor, initializes self.fs with BytesIO.
+
+        :param fp: `BytesIO` / `IOBase`: Open file, either in-memory
+                   or already open handle.
+        :param encoding: `str`: Valid Python standard encoding.
+        :param offset: `int`: Offset from file start to filesystem
+                       start in bytes.
+        :param preserve_case: `bool`: By default 8DOT3 filenames do not
+                              support casing. If preserve_case is set to
+                              `True`, it will create an LFN entry if the
+                              casing does not conform to 8DOT3.
+        :param utc: `bool`: Store timestamps in UTC rather than the local time.
+                    This applies to dentry creation, modification and last
+                    access time.
+        """
+        super(PyFatFS, self).__init__()
+        self.preserve_case = preserve_case
+        self.fs = PyFat(encoding=encoding, offset=int(offset))
+        self.fs.set_fp(fp)
+
+        if utc:
+            self.tz = datetime.timezone.utc
+        else:
+            self.tz = datetime.datetime.now(datetime.timezone.utc)
+            self.tz = self.tz.astimezone().tzinfo
